/**
 * <p>Title: Informatik IIb - bungen</p>
 * <p>Description: Endlicher deterministischer Automat</p>
 * <p>Company: Fachhochschule Jena</p>
 * @author Oliver Jack
 * @version 1.0
 */

/*
 * Liest aus einer Textdatei
 * einen endlichen deterministischen Automaten ein.
 * Stellt Methode action(String) zum Betrieb des Automaten bereit.
 * 
 * Textdateiformat:
 * # Kommentar
 * Zustand > Transition > neuer Zustand
 * ...
 * 
 * Vereinbarung:
 * mit
 * _start > new > Initialzustand
 * wird der Initialzustand beschrieben
 */

import java.io.*;
import java.util.*;

public class FSM {

	private String state = new String("_start");
	private String message = new String();

	// key: state; value: <transition, new state>
	private HashMap<String, HashMap<String, String>> states = new HashMap<String, HashMap<String, String>>();

	FSM(String fileName) throws IOException {
		String transition;
		String token[] = new String[3]; // parse line
		boolean seenInitialState = false;
		BufferedReader f;
		f = new BufferedReader(new FileReader(fileName));

		// Read fsm
		while ((transition = f.readLine()) != null) {
			// Ignore comments and empty lines
			if (transition.trim().startsWith("#")
					|| transition.trim().isEmpty()) {
				continue;
			}
			token = transition.split(">");
			if (token.length != 3)
				throw (new IOException("Syntaxfehler: " + transition.toString()));
			// we have a transition
			for (int i = 0; i < 3; ++i)
				token[i] = token[i].trim();
			// sort out initial state
			if (token[0].equals("_start") && token[1].equals("new")) {
				if (seenInitialState)
					throw (new IOException("Redefinition of initial state: "
							+ state + "->" + token[2]));
				state = token[2];
				seenInitialState = true;
			}
			// new state
			if (states.get(token[0]) == null) {
				states.put(token[0], new HashMap<String, String>());
				states.get(token[0]).put(token[1], token[2]);
			} else {
				// new transition for existing state
				if (states.get(token[0]).get(token[1]) == null) {
					states.get(token[0]).put(token[1], token[2]);
				} else {
					if (states.get(token[0]).get(token[1]).equals(token[2])) {
						System.out.println("Warning: transition " + token[0]
								+ " > " + token[1] + " > " + token[2]
								+ " multiply defined");
					} else {
						throw (new IOException("Transition " + token[1]
								+ " for state " + token[0] + " redefined"));
					}
				}
			}
		}
		if (!seenInitialState) {
			throw (new IOException("No initial state defined"));
		}
		f.close();
	}

	public boolean action(String sAction) {
		String newState = new String();
		if (states.get(state) != null) {
			// there is such a transition for this state
			if ((newState = states.get(state).get(sAction)) != null) {
				message = sAction;
				state = newState;
				return true;
			} else {
				message = sAction + " (in Zustand " + state + " nicht mglich)";
				return false;
			}
		} else {
			message = sAction + " (interner Fehler, Zustand verloren)";
			return false;
		}
	}

	public void print() {
		String iState = new String();
		@SuppressWarnings({ "rawtypes", "unchecked" })
		List stateKeys = new ArrayList(states.keySet());
		@SuppressWarnings("rawtypes")
		Iterator s = stateKeys.iterator();
		System.out.println("Automat");
		System.out.println(" Zustnde: " + stateKeys);
		System.out.println(" Transitionen:");
		while (s.hasNext()) {
			iState = (String) s.next();
			System.out.println("  " + iState + ": " + states.get(iState));
		}
	}

	public String toString() {
		return ("Zustand " + state + " durch " + message);
	}
}
